# pylint: disable=import-error
# mypy: ignore-errors
"""
Cache repository implementation.
"""
import asyncio
import json
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar

from typing_extensions import ParamSpec

from .repository import Repository

_A = TypeVar("_A")
_U = TypeVar("_U")
_R = TypeVar("_R")
_Id = TypeVar("_Id")
_I = TypeVar("_I")
_WRapperIn = TypeVar("_WRapperIn")
_WrapperOut = TypeVar("_WrapperOut")
_Params = ParamSpec("_Params")


class CacheRepository(
    Repository[_Id, _A, _U, _R, _I],
    Generic[_Id, _A, _U, _R, _I],
):
    """A cached repository implementation.

    This implements caching for an underlying repository, provided in the constructor.

    For simplisity, the implementation relies in the functool's caching functionality.

    Note that modify operations are not cached and clear the caches.
    """

    def __init__(self, repository: Repository[_Id, _A, _U, _R, _I]) -> None:
        super().__init__()
        self.repository = repository
        self._cache: Dict[str, Any] = {}

    def clear_cache(self):
        """Clears the repository-level cache."""

        self._cache.clear()

    async def add(self, payload: _A, **kwargs: Any) -> _I:
        return await self.repository.add(payload)

    async def get_list(
        self,
        *,
        offset: Optional[int] = None,
        size: Optional[int] = None,
        **query_filters: Any,
    ) -> List[_I]:
        data: asyncio.Task[List[_I]] = self._get_or_cache(
            self.repository.get_list,
            "list",
            asyncio.create_task,
            offset=offset,
            size=size,
            **query_filters,
        )

        return await data

    def _get_or_cache(
        self,
        method: Callable[_Params, _WRapperIn],
        prefix: str,
        wrapper: Callable[[_WRapperIn], _WrapperOut],
        *args: _Params.args,
        **kwargs: _Params.kwargs,
    ) -> _WrapperOut:
        cache_key = self._gen_cache_key(prefix, *args, **kwargs)
        data = self._cache.get(cache_key)
        if data is None:
            original_data = method(*args, **kwargs)
            data = wrapper(original_data)
            self._cache[cache_key] = data
        return data

    def _gen_cache_key(self, prefix: str, *args: Any, **kwargs: Any) -> str:
        body = json.dumps({**kwargs, "__args": args})
        return f"{prefix}:{body}"

    async def get_count(self, **query_filters: Any) -> int:
        data: asyncio.Task[int] = self._get_or_cache(
            self.repository.get_count,
            "count",
            asyncio.create_task,
            **query_filters,
        )

        return await data

    async def get_by_id(self, item_id: _Id, **kwargs: Any) -> _I:
        data: asyncio.Task[_I] = self._get_or_cache(
            self.repository.get_by_id,
            "get_by_id",
            asyncio.create_task,
            item_id,
            **kwargs,
        )

        return await data

    async def update(self, item_id: _Id, payload: _U, **kwargs: Any) -> _I:
        result = await self.repository.update(item_id, payload, **kwargs)
        self.clear_cache()
        return result

    async def replace(self, item_id: _Id, payload: _R, **kwargs: Any) -> _I:
        result = await self.repository.replace(item_id, payload, **kwargs)
        self.clear_cache()
        await self.get_by_id(item_id)
        return result

    async def remove(self, item_id: _Id, **kwargs: Any):
        await self.repository.remove(item_id, **kwargs)
        self.clear_cache()



"""
Composite repository implementation.

This implements a composite pattern for repositories.
"""
from typing import Any, Generic, List, Optional, TypeVar

from .mapper import Mapper
from .repository import Repository
from .utils import merge_dicts

_MI = TypeVar("_MI")
_MA = TypeVar("_MA")
_MU = TypeVar("_MU")
_MR = TypeVar("_MR")
_MId = TypeVar("_MId")
_A = TypeVar("_A")
_U = TypeVar("_U")
_R = TypeVar("_R")
_I = TypeVar("_I")
_Id = TypeVar("_Id")


class MappedRepository(
    Repository[_MId, _MA, _MU, _MR, _MI],
    Generic[
        _MId,
        _MA,
        _MU,
        _MR,
        _MI,
        _Id,
        _A,
        _U,
        _R,
        _I,
    ],
):
    """Mapped data for repositories.

    This implements the repository interface by leveraging mappers between payloads,
    item and id for an underlying repository implementation.
    """

    def __init__(
        self,
        repository: Repository[_Id, _A, _U, _R, _I],
        *,
        id_mapper: Mapper[_MId, _Id],
        create_mapper: Mapper[_MA, _A],
        update_mapper: Mapper[_MU, _U],
        replace_mapper: Mapper[_MR, _R],
        item_mapper: Mapper[_I, _MI],
        **query_filters: Any,
    ) -> None:
        """Initialize the `MappedRepository` instance.

        Args:
            id_mapper: A mapper to transform the item ID.
            create_mapper: A mapper to add new items.
            update_mapper: Mapper to update an item.
            replace_mapper: Maps the replace payload.
            item_mapper: The item mapper from the repository implementation.
            repository: The underlying repository implementations.
        """
        super().__init__()
        self.repository = repository
        self.item_mapper = item_mapper
        self.id_mapper = id_mapper
        self.create_mapper = create_mapper
        self.update_mapper = update_mapper
        self.replace_mapper = replace_mapper
        self._filters = query_filters

    async def add(self, payload: _MA, **kwargs: Any) -> _MI:
        """Add a new item.

        Args:
            payload (_MA): The payload to use

        Returns:
            _MI: The newly created item
        """
        return await self.map_item(
            await self.repository.add(self.create_mapper(payload), **kwargs)
        )

    async def update(self, item_id: _MId, payload: _MU, **kwargs: Any) -> _MI:
        """Update an item.

        Args:
            item_id (_MId): The item ID to change
            payload (_MU): The data to update with

        Returns:
            _MI: The updated item
        """
        return await self.map_item(
            await self.repository.update(
                self.id_mapper(item_id),
                self.update_mapper(payload),
                **merge_dicts(kwargs, self._filters),
            )
        )

    async def get_by_id(self, item_id: _MId, **kwargs: Any) -> _MI:
        """Retrieve an item by it's ID.

        Args:
            item_id (_MId): The ID of the item to retrieve

        Returns:
            _MI: The item
        """
        return await self.map_item(
            await self.repository.get_by_id(
                self.id_mapper(item_id), **merge_dicts(kwargs, self._filters)
            )
        )

    async def replace(self, item_id: _MId, payload: _MR, **kwargs: Any) -> _MI:
        """Replace an item in the underlying store.

        Args:
            item_id (_MId): The item ID to replace
            payload (_MR): The data to replace with

        Returns:
            _MI: The new item
        """
        return await self.map_item(
            await self.repository.replace(
                self.id_mapper(item_id),
                self.replace_mapper(payload),
                **merge_dicts(kwargs, self._filters),
            )
        )

    async def get_count(self, **query_filters: Any) -> int:
        """Retrieve a count of items.

        Returns:
            int: The item count
        """
        return await self.repository.get_count(
            **merge_dicts(query_filters, self._filters)
        )

    async def get_list(
        self,
        *,
        offset: Optional[int] = None,
        size: Optional[int] = None,
        **query_filters: Any,
    ) -> List[_MI]:
        """Retrieve a list of items from the undeerlying repository.

        Args:
            offset (Optional[int], optional): The list offset. Defaults to None.
            size (Optional[int], optional): The list size. Defaults to None.

        Returns:
            List[_MI]: A list of items
        """
        return await self.map_item_list(
            await self.repository.get_list(
                offset=offset,
                size=size,
                **merge_dicts(query_filters, self._filters),
            )
        )

    async def remove(self, item_id: _MId, **kwargs: Any):
        """Remove an item from the underlying repository.

        Args:
            item_id (_MId): The item ID to be removed.
        """
        await self.repository.remove(
            self.id_mapper(item_id),
            **merge_dicts(kwargs, self._filters),
        )

    async def map_item(self, item: _I) -> _MI:
        """Transform an item.

        Args:
            item (_I): The original item

        Returns:
            _MI: The transformed item
        """
        return self.item_mapper(item)

    async def map_item_list(self, item_list: List[_I]) -> List[_MI]:
        """Map an item list to the correct representation.

        Args:
            item_list (List[_I]): The original item list

        Returns:
            List[_MI]: The mapped item list
        """
        return [self.item_mapper(item) for item in item_list]



import typing


class MergeError(Exception):
    def __init__(self, location: typing.List[typing.Any], message: str) -> None:
        super().__init__(f"Error at {'->'.join(map(str, location))}: {message}")


def merge_dicts(*dicts: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
    """Merge two or more dicts into new one.

    Returns:
        typing.Dict[str, typing.Any]: The resulting dictionary.

    A simple merge example:
    >>> merge_dicts({'a': 1}, {'b': 2})
    {'a': 1, 'b': 2}
    >>>

    For a nested merge:
    >>> dict_a={'x': {'a': 1}}
    >>> dict_b={'x': {'b': 1}}
    >>> result = merge_dicts(dict_a, dict_b)
    >>> result == {'x': {'a': 1, 'b': 1}}
    True
    >>>
    """
    result = _merge_dicts(*dicts)

    return result


def _merge_dicts(  # noqa C901
    *dicts: typing.Dict[str, typing.Any],
    location: typing.Optional[typing.List[typing.Any]] = None,
) -> typing.Dict[str, typing.Any]:
    result: typing.Dict[str, typing.Any] = {}
    current_key = []
    if location is not None:
        current_key += location

    for i, source in enumerate(dicts):
        current_key.append(i)
        for key, value in source.items():
            current_key.append(key)
            current = result.get(key, None)

            if isinstance(current, dict):
                if not isinstance(value, dict):
                    raise MergeError(
                        current_key.copy(),
                        "Type mismatch: The right value is not a dict.",
                    )
                result[key] = _merge_dicts(current, value, location=current_key)
            elif isinstance(current, list):
                if isinstance(value, (set, tuple, list)):
                    result[key] = current + [*value]
                else:
                    result[key] = current + [value]
            else:
                result[key] = value

            current_key.pop()

        current_key.pop()

    return result



"""
HTTP repository implementation.

This module implements a subclass of the `generic_repository.repository.Repository`
class with http actions.
"""
import cgi
from http import HTTPStatus
from typing import Any, Dict, List, Optional, cast

import httpx  # pylint: disable=import-error

from .exceptions import InvalidPayloadException, ItemNotFoundException
from .mapper import LambdaMapper, Mapper
from .repository import Repository


class HttpRepository(Repository[str, Any, Any, Any, Any]):
    """An http repository.

    To use it, you must provide an httpx async client.

    For example:
    >>> import asyncio
    >>> from httpx import AsyncClient
    >>> client = AsyncClient(base_url='https://jsonplaceholder.typicode.com')
    >>> repo = HttpRepository(client, base_url='/posts')
    >>> repo
    <generic_repository.http.HttpRepository object at ...>
    >>>
    """

    def __init__(
        self,
        client: httpx.AsyncClient,
        *,
        base_url: Optional[str] = None,
        request_params: Optional[Dict[str, Any]] = None,
        count_mapper: Optional[Mapper[Any, int]] = None,
        list_mapper: Optional[Mapper[Any, List[Any]]] = None,
        add_slash: bool = True,
    ) -> None:
        super().__init__()
        self.client = client
        self._base_url = base_url
        self._request_params = request_params
        self.count_mapper = count_mapper or LambdaMapper(
            lambda x: len(cast(List[Any], x))
        )

        self.list_mapper = list_mapper or LambdaMapper(lambda x: cast(List[Any], x))
        self.add_slash = add_slash

    @property
    def base_url(self):
        """Return the base URL of the resource.

        You can override this to customize the behavior.
        """
        if self._base_url is not None:  # pragma: nocover
            return self._base_url
        return ""

    @property
    def request_params(self):
        """Return additional request parameters.

        This is the easyest way to customize the request headers and query parameters.
        """
        params = {}

        if self._request_params is not None:  # pragma nocover
            params.update(self._request_params)

        return params

    async def process_response(self, response: httpx.Response) -> Any:
        """Processes the response coming from the HTTP server.

        Args:
            response: The response result.

        Returns:
            Any: The response raw data.
        """
        self._ensure_success(response)

        if response.status_code == HTTPStatus.NO_CONTENT:
            return None
        elif (
            response.status_code >= 300 and response.status_code < 400
        ):  # pragma nocover
            raise RuntimeError("A redirect found.")

        try:
            if self.is_json(response):
                return response.json()
            else:  # pragma nocover
                # Return the response text representation:
                return response.text
        finally:
            await response.aclose()

    @classmethod
    def _ensure_success(cls, response):
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as ex:
            if response.status_code == HTTPStatus.NOT_FOUND:
                raise ItemNotFoundException("Cannot find the remote resource.") from ex
            elif response.status_code in (
                HTTPStatus.BAD_REQUEST,
                HTTPStatus.UNPROCESSABLE_ENTITY,
            ):
                raise InvalidPayloadException() from ex
            else:
                raise

    @classmethod
    def is_json(cls, response: httpx.Response) -> bool:
        """Checks if the response contains JSON data.

        Args:
            response: The response to be checked.

        Returns:
            bool: `True` if the response is a JSON document, false otherwise.
        """
        content_type, _ = cgi.parse_header(response.headers.get("content-type"))
        _main_type, sub_type = content_type.split("/")
        is_json = sub_type.startswith("json")
        return is_json

    async def get_by_id(self, item_id: str, **kwargs: Any) -> Any:
        """Return a resource by it's ID.

        Args:
            item_id: The ID of the resource to retrieve.

        Returns:
            Any: The resource as Json.
        """
        return await self.process_response(
            await self.client.get(
                self.get_id_url(item_id), **self._merge_params(kwargs)
            )
        )

    async def _get_list(self, *, offset=None, size=None, **query_filters):
        params = {}
        if size is not None:
            params.update(size=size)
        if offset is not None:
            params.update(offset=offset)

        request_params = self._merge_params(query_filters, params)
        return await self.process_response(
            await self.client.get(self.list_url, **request_params)
        )

    def _merge_params(
        self, query_filters: Dict[str, Any], params: Optional[Dict[str, Any]] = None
    ):
        if params is None:
            params = {}
        request_params = {
            "params": {
                **self.request_params.get("params", {}),
                **query_filters.get("params", {}),
                **params,
            },
            "headers": {
                **self.request_params.get("headers", {}),
                **query_filters.get("headers", {}),
            },
            "cookies": {
                **self.request_params.get("cookies", {}),
                **query_filters.get("cookies", {}),
            },
        }

        return request_params

    @property
    def list_url(self):
        """Retrieve the real base_url.

        Returns:
            str: The base URL built from the passed base url in the constructor.

        >>> base_url = 'https://example.com/resource'
        >>> repo = HttpRepository(
        ...     httpx.AsyncClient(),
        ...     base_url=base_url,
        ...     add_slash=False,
        ... )
        >>> repo.list_url
        'https://example.com/resource'
        >>> repo2 = HttpRepository(
        ...     repo.client, base_url=base_url, add_slash=True
        ... )
        >>> repo2.list_url
        'https://example.com/resource/'
        >>>
        """
        list_url = self.base_url
        if self.add_slash:
            list_url = f"{list_url}/"
        return list_url

    async def get_list(self, **query_filters):
        """Return a list of items.

        Args:
            offset (int, optional): Where to start from.. Defaults to None.
            size (int, optional): How many items to retrieve. Defaults to None.

        Returns:
            list: Items from the remote source.

        """
        return self.list_mapper(await self._get_list(**query_filters))

    async def get_count(self, **query_filters: Any) -> int:
        """Return how many items the remote source have in it's database.

        Returns:
            int: How many items.
        """
        return self.count_mapper(await self._get_list(**query_filters))

    async def add(self, payload: Any, **kwargs: Any) -> Any:
        """Add a new item to the remote resource.

        Args:
            payload: A payload to use as the data source.

        Returns:
            Any: The newly created item.
        """
        extra_data = kwargs.pop("extra_data", None)
        if extra_data is not None:
            if not isinstance(payload, dict):
                raise ValueError("Cannot update the payload with the extra data.")
            elif not isinstance(extra_data, dict):
                raise ValueError("Invalid extra data provided.")
            else:
                # The real data takes presedence:
                payload = {**extra_data, **payload}

        return await self.process_response(
            await self.client.post(
                self.list_url, json=payload, **self._merge_params(kwargs, extra_data)
            )
        )

    async def update(self, item_id: str, payload: Any, **kwargs: Any) -> Any:
        """Update the remote resource.

        Args:
            item_id: The resource ID.
            payload: The payload to send. Must be json-compatible.

        Returns:
            Any: Json-compatibel updated data.
        """
        return await self.process_response(
            await self.client.patch(
                self.get_id_url(item_id), json=payload, **self._merge_params(kwargs)
            )
        )

    def get_id_url(self, item_id):
        """Returns the URL for an item id.

        Args:
            item_id: The item ID to get URL for.

        Returns:
            str: The URL of the item ID.
        """
        return f"{self.base_url}/{item_id}"

    async def replace(self, item_id: str, payload: Any, **kwargs: Any):
        """Send a replace request.

        Args:
            item_id: The resource ID to replace.
            payload (Any): The new data for the resource.

        Returns:
            Any: The newly updated resource.
        """
        return await self.process_response(
            await self.client.put(self.get_id_url(item_id), json=payload)
        )

    async def remove(self, item_id: str, **kwargs: Any):
        """Remove the specified remote resource.

        Args:
            item_id: The resource ID to remove.
        """
        await self.process_response(
            await self.client.delete(
                self.get_id_url(item_id), **self._merge_params(kwargs)
            )
        )



# flake8: noqa F401

from .cached import CacheRepository
from .composition import MappedRepository
from .exceptions import CrudException, InvalidPayloadException, ItemNotFoundException
from .mapper import ConstructorMapper, LambdaMapper, Mapper, ToFunctionArgsMapper
from .repository import Repository

GenericBaseRepository = Repository

__all__ = [
    # Base classes
    "Repository",
    # Cache-based implementations:
    "CacheRepository",
    # Composition:
    "MappedRepository",
    "ConstructorMapper",
    "LambdaMapper",
    "Mapper",
    "ToFunctionArgsMapper",
    # Exceptions:
    "CrudException",
    "InvalidPayloadException",
    "ItemNotFoundException",
]

try:
    from .database import (
        DatabaseRepository,
        SqlalchemyMappedRepository,
        SqlalchemyModelRepository,
    )
except ImportError:  # pragma nocover
    pass
else:
    __all__ += [
        "DatabaseRepository",
        "SqlalchemyMappedRepository",
        "SqlalchemyModelRepository",
    ]

try:
    from .http import HttpRepository
except ImportError:  # pragma nocover
    pass
else:
    __all__ += ["HttpRepository"]

try:
    from .pydantic import PydanticDictMapper, PydanticObjectMapper
except ImportError:  # pragma nocover
    pass
else:
    __all__ += ["PydanticDictMapper", "PydanticObjectMapper"]
